/*
vim:ts=4:sw=4:tw=0
*/

#include <stdio.h>
#include <fcntl.h>
#include <malloc.h>
#include <sys/stat.h>
#include <linux/soundcard.h>

#include "play.h"

int header_check(wave_header *head, int fd, header_data *hd) {
	unsigned p;
	int foo;

	if ( strncmp(head->main_chunk,"RIFF",4) ) {
		err(0,"Not in RIFF format.");
		return 0;
	}

	if ( strncmp(head->chunk_type,"WAVE",4) ) {
		err(0,"Not a WAVE type.");
		return 0;
	}
		
	if (hd->debug_mode)
		fprintf(stderr, "**DEBUG: Pre-correction:\n"
						"  '%.4s'[%d],  '%.4s',  '%.4s'[%d],  Format= %s(%s),  "
						"%u Hz,\n  Data per sec: %u, Bytes per sec: %hu,  Bytes "
						"per sample: %hu, '%.4s'[%d]\n",
						head->main_chunk, head->length,
						head->chunk_type, head->sub_chunk, head->length_chunk,
						(head->format == 1?"PCM":"Unknown"),
						(head->modus == 1?"Mono":"Stereo"),
						head->sample_fq, head->byte_p_sec,
						head->byte_p_spl, head->bit_p_spl,
						head->data_chunk, head->data_length);
	
	/* Scan for fmt-chunk, then re-read it */
	if (strncmp(head->sub_chunk, "fmt ", 4)) {
		lseek(fd, 0, SEEK_SET);
		while(strncmp(head->sub_chunk, "fmt ", 4)) {
			lseek(fd, -2, SEEK_CUR);		/* Catch off-by-two errors */
			if (4 != read(fd, &head->sub_chunk, 4)) {
				err(1,"Error scanning file for fmt-chunk!"); return 0;
			}
		}
		if (4 != read(fd, &head->length_chunk, 4)) {
			err(1,"Found fmt-chunk, but couldn't read length field!"); return 0;
		}
		foo = sizeof(wave_header) - offsetof(wave_header,format);
		if (foo != read(fd, &head->format, foo)) {
			err(1,"Reading fmt-chunk failed!"); return 0;
		}
	}

	if (hd->debug_mode)
		fprintf(stderr, "**DEBUG: Post-fmt-chunk correction, Pre-data-chunk correction:\n"
						"  '%.4s'[%d],  '%.4s',  '%.4s'[%d],  Format= %s(%s),  "
						"%u Hz,\n  Data per sec: %u, Bytes per sec: %hu,  Bytes "
						"per sample: %hu, '%.4s'[%d]\n",
						head->main_chunk, head->length,
						head->chunk_type, head->sub_chunk, head->length_chunk,
						(head->format == 1?"PCM":"Unknown"),
						(head->modus == 1?"Mono":"Stereo"),
						head->sample_fq, head->byte_p_sec,
						head->byte_p_spl, head->bit_p_spl,
						head->data_chunk, head->data_length);
	/* Do we really have a data-chunk? Or is there some other junk there? */
	if (strncmp(head->data_chunk, "data", 4)) {
		/* Ah, there wasn't any data-chunk! Well, search for it then! (Of
		 * course, we could have done this much quicker, by calculating where
		 * the data-chunk would be most likely to begin, but then we would have
		 * to read more data into the header-buffer etc. etc... Nah, we'll do
		 * it the simple and dirty way!
		 */
		lseek(fd, -2, SEEK_CUR);		 /* Go back 2 bytes, so we won't miss anything. */
		while(strncmp(head->data_chunk, "data", 4)) {
			lseek(fd, -2, SEEK_CUR);		/* Catch off-by-two errors */
			if (4 != read(fd, &head->data_chunk, 4)) {
				err(1,"Error scanning file for data-chunk!");
				return 0;
			}
		}
		/* So far, so good. Now try to read the length of the data-chunk! */
		if (4 != read(fd, &head->data_length, 4)) {
			err(1,"Found data-chunk, but couldn't read length field!"); return 0;
		}
		/* Alright! We've found the data-chunk, and fooled the header into
		 * believeing that everyting is alright! My, my. Will the cleverness of
		 * Mr. Myrvang never end? :) */
	}

	/* One final check, this time to ensure that the data_length is within sane
	 * limits, i.e. the length of the file... */
	if (head->data_length >= head->length) {
		err(0, "Gee! Sample-data sequence is longer than the file! Truncating!");
		head->data_length = head->length - lseek(fd, 0, SEEK_CUR);
	}

	if (hd->debug_mode)
		fprintf(stderr, "**DEBUG: Post-correction:\n"
						"  '%.4s'[%d],  '%.4s',  '%.4s'[%d],  Format= %s(%s),  "
						"%u Hz,\n  Data per sec: %u, Bytes per sec: %hu,  Bytes "
						"per sample: %hu, '%.4s'[%d]\n",
						head->main_chunk, head->length,
						head->chunk_type, head->sub_chunk, head->length_chunk,
						(head->format == 1?"PCM":"Unknown"),
						(head->modus == 1?"Mono":"Stereo"),
						head->sample_fq, head->byte_p_sec,
						head->byte_p_spl, head->bit_p_spl,
						head->data_chunk, head->data_length);


/* Wrong length difference is no longer a fatal error, and the handling here
 * was wrong anyway (fixed data-positions are nonexistant with IFF-like files,
 * they _have_ to be parsed!), so we can just skip this check. */
#if 0
	p = head->length - head->data_length; if ( (p != 36) && (p != 37) ) {
		err(0,"Wrong length difference. "); fprintf(stderr, "That is,
		head->length - head->data_length = %d\n", p); return 0; }
#endif


	if ( strncmp(head->sub_chunk,"fmt ",4) ) {
		err(0,"Not a WAVE format.");
		return 0;
	}

	if ( (head->sample_fq * head->byte_p_spl) != head->byte_p_sec ) 
		err(0,"Incorrect sampling rate.");

	return 1;
}

/*********************************************************************
 * Record or Play samples :
 *
 * mode = PLAY / RECORD
 * data = command_line options
 *********************************************************************/
void recplay(char *filename,int mode,header_data data,int ipc) {
	int tmp;
	int count;			/* Total # of bytes */
	int cnt;			/* bytes for this I/O */
	int n;				/* Bytes read/written */
	int omode = (mode == PLAY) ? O_WRONLY : O_RDONLY;	
	int um = 0666;			/* Current umask() value */
	int fd = -1;			/* read/write file descriptor */
	int audio = -1;			/* /dev/dsp audio file descriptor */
	int abuf_size;			/* audio buffer size for I/O */
	char *iobuf;			/* allocated I/O buffer */
	char tbuf[80];			/* Temporary buffer */
	wave_header whdr;		/* header to/from file */

	if (data.debug_mode)
		fprintf(stderr, "\n**DEBUG: filename = \"%s\"\n", filename);

	/*************************************************************
	 * Find out what are umask setting is :
	 *************************************************************/

	um = umask(0666);
	umask(um);		/* Restore umask */

	/*************************************************************
	 * Lock the device :
	 *************************************************************/

	if ( ipc >= 0 ) {
		if ( LockDSP(ipc,mode==PLAY?0:1) ) {
			ipc = -1;		/* Don't unlock now */
			err(1,mode==PLAY
				? "Locking audio device for play."
				: "Locking audio device for record.");
			goto errxit;
		}
	}

	if ( mode == PLAY ) {
		/*****************************************************
		 * Play a sample :
		 *****************************************************/

		if ( filename == NULL ) {
			fd = 0;
			filename = "stdin";
		} else if ( (fd = open(filename,O_RDONLY,0)) < 0 ) {
			sprintf(tbuf,"Error opening %s",filename);
			err(1,tbuf);
			goto xit;
		}

		if ( (n = read(fd,&whdr,sizeof(whdr))) < 0 ) {
			sprintf(tbuf,"Reading \"%s\". Skipping...",filename);
			err(1,tbuf);
			goto xit;

		} else if ( n == 0 )
			goto xit;	/* Perhaps its /dev/null */

		else if ( n < sizeof(whdr) ) {
			err(0,"Short header: can't be wav file.");
			goto xit;
		}

		if (0 != data.noplay)
			fprintf(stderr, "Checking: %s\n", filename);
		
		if ( !header_check(&whdr, fd, &data) )
			goto xit; 
		
		if ( !data.force_stereo )
			data.stereo = (whdr.modus == 2);

		if ( !data.force_speed )
			data.speed = whdr.sample_fq;

		if ( !data.force_sample_size ) 
			data.sample_size = whdr.bit_p_spl;
			
		/*****************************************************
		 * set upper border accordingly
		 *****************************************************/
			
		if ( data.time_limit == 0.0 ) 
			data.upper_border = whdr.data_length;
		else	{
			if (!data.force_speed)
				data.upper_border = data.time_limit * whdr.sample_fq;
			else	data.upper_border = data.time_limit * data.speed;
		
			if ( data.stereo )
				data.upper_border *= 2;
			if ( data.sample_size != 8 )
				data.upper_border *= 2;
		}
	} else	{				/* mode = RECORD */
		/*****************************************************
		 * Record a sample :
		 *****************************************************/

		if ( filename == NULL ) {
			fd = 1;
			filename = "stdout";
		} else	{
			remove(filename);
			if ( (fd = open(filename,O_WRONLY|O_CREAT|O_EXCL,0666&~um)) < 0 ) {
				sprintf(tbuf,"Opening %s for write.",filename);
				err(1,tbuf);
				goto errxit;
			}
		}

		/*****************************************************
		 * Create header :
		 *****************************************************/

		strncpy(whdr.main_chunk,"RIFF",4);
		strncpy(whdr.chunk_type,"WAVE",4);
		strncpy(whdr.sub_chunk,"fmt ",4);
		strncpy(whdr.data_chunk,"data",4);

		whdr.length_chunk = 16;
		whdr.format = 1;
		whdr.modus = data.stereo ? 2 : 1;		
		whdr.sample_fq = data.speed;

		if ( !data.stereo ) {
			/*********************************************
			 * Mono :
			 *********************************************/
			if ( data.sample_size == 8 )
				whdr.byte_p_spl = 1;
			else 	whdr.byte_p_spl = 2;
		} else	{
			/*********************************************
			 * Sterio :
			 *********************************************/
			if ( data.sample_size == 8 )
				whdr.byte_p_spl = 2;
			else	whdr.byte_p_spl = 4;
		}

		whdr.byte_p_sec = data.speed * whdr.byte_p_spl;
		whdr.data_length = whdr.byte_p_sec * data.time_limit;

		whdr.bit_p_spl = data.sample_size;
		whdr.length = whdr.data_length + 36;
		data.upper_border = data.time_limit * data.speed
			* whdr.byte_p_spl;
		data.lower_border = 0;

		if ( write(fd,&whdr,sizeof(whdr)) != sizeof(whdr) ) {
			err(1,"Writing WAVE header.");
		   	return;
		}
	}

	count = data.upper_border - data.lower_border;

	if ( !data.quiet_mode ) {
		if ( mode == PLAY ) {
			if (!data.noplay) {
				fprintf(stderr,"Playing: %s", filename);
			}
		}
		else	fprintf(stderr,"Recording: %s",filename);

		if ( data.time_limit == 0.0 ) {
			if (!data.noplay) {
				fprintf(stderr," (in full)\n");
			}
		}
		else	fprintf(stderr," (%5f secs)\n",data.time_limit);

		if ( data.stereo ) 
			fprintf(stderr,"Stereo");
		else	fprintf(stderr,"Mono");
		fprintf(stderr,", rate %u Hz, %u bits.\n", 
			data.speed, data.sample_size);
	}

	if (0 != data.noplay)
		goto xit;

	if ( (audio = open(AUDIODEV,omode,0)) < 0 ) {
		err(1,"Error opening audio-device");
		goto errxit;
	}

	if ( ioctl(audio,SNDCTL_DSP_GETBLKSIZE,&abuf_size) < 0 ) {
		err(1,"Obtaining dsp buffer size.");
		goto errxit;
	}

	if ( abuf_size < 4096 || abuf_size > 65536 ) {
		sprintf(tbuf,"Invalid audio buffer size (%d bytes).",(int)abuf_size);
		err(1,tbuf);
		goto errxit;
	} 

	if ( (iobuf = malloc(abuf_size)) == NULL ) {
		err(1,"Unable to allocate audio buffer.");
		goto errxit;
	}

        tmp = data.sample_size;
        if ( ioctl(audio, SNDCTL_DSP_SAMPLESIZE, &data.sample_size) < 0 ) {
        	err(1,"Unable to set audio sample size.");
		goto errxit;
	}

	if ( tmp != data.sample_size ) { /* Is this test needed? */
		err(0,"Unable to set sample size");
		goto errxit;
	}

	if ( ioctl(audio, SNDCTL_DSP_STEREO, &data.stereo) < 0 ) {
		err(1,"Unable to set stereo/mono.");
		goto errxit;
	}		
      
	if ( ioctl(audio,SNDCTL_DSP_SPEED,&data.speed) < 0 ) {
		err(1,"Unable to set audio sampling rate.");
		goto errxit;
	}

	if ( mode == PLAY ) {
		/*****************************************************
		 * Play :
		 *****************************************************/

		while( count > 0 ) {
			cnt = count > abuf_size ? abuf_size : count;

			if ( (n = read(fd,iobuf,cnt)) < 0 ) {
				err(1,"Reading audio sample.");
				goto xit;
			}

			if ( n == 0 ) {
				err(0,"Unexpected EOF reading samples.");
				goto xit;
			}

			if ( write(audio,iobuf,n) != n ) {
				err(1,"Writing audio device.");
				goto xit;
			}

			count -= n;
		}
			
	} else	{
		/*****************************************************
		 * Record :
		 *****************************************************/

		while ( count > 0 ) {
			cnt =  count > abuf_size ? abuf_size : count;

			if ( (n = read(audio,iobuf,cnt)) < 0 ) {
				err(1,"Reading from audio device.");
				goto errxit;
			}

			if ( write(fd,iobuf,n) != n ) {
				err(1,"Writing to output.");
				goto errxit;
			}

			count -= n;
		}
	}

	/*************************************************************
	 * Normal Exit :
	 *************************************************************/

xit:	close(fd);
	close(audio);
	if ( ipc >= 0 ) {
		/*****************************************************
		 * Unlock so other processes can use it now :
		 *****************************************************/
		if ( UnlockDSP(ipc,mode==PLAY?0:1) ) {
			err(1,mode==PLAY
				? "Unlocking for audio play."
				: "Unlocking for audio record.");
		}
	}
	return;

	/*************************************************************
	 * Error Exit :
	 *************************************************************/

errxit:	if ( fd > 0 )
		close(fd);
	close(audio);
	if ( ipc >= 0 )
		UnlockDSP(ipc,mode==PLAY?0:1);	/* Best effort basis */
	exit(1);
}
